import base64
from binascii import Error
from collections import deque, namedtuple
from contextlib import suppress
from keyword import kwlist
from pickle import UnpicklingError
from pickle import dumps as generic_dumps
from pickle import loads as generic_loads
from typing import ClassVar

from kalib.classes import Nothing
from kalib.descriptors import cache
from kalib.functions import to_ascii, to_bytes
from kalib.importer import optional, required, sort
from kalib.internals import (
    Who,
    class_of,
    is_collection,
    is_function,
    is_mapping,
    sourcefile,
)
from kalib.loggers import Logging
from kalib.monkey import Monkey
from kalib.text import Str

BACKENDS = {b'json': 'orjson', b'ujson': 'orjson'}
DEADBEEF = b'\xDE\xAD\xBE\xEF'

serializers = {}
logger = Logging.get(__name__)


try:
    json = required(BACKENDS[b'json'], quiet=True)

    from orjson import (
        OPT_INDENT_2,
        OPT_NAIVE_UTC,
        OPT_NON_STR_KEYS,
        OPT_SERIALIZE_DATACLASS,
        OPT_SERIALIZE_NUMPY,
        OPT_SERIALIZE_UUID,
        OPT_SORT_KEYS,
        OPT_STRICT_INTEGER,
        JSONDecodeError,
    )

    OPT_JSON_FLAGS = (
        OPT_NAIVE_UTC |
        OPT_NON_STR_KEYS |
        OPT_SERIALIZE_DATACLASS |
        OPT_SERIALIZE_NUMPY |
        OPT_SERIALIZE_UUID |
        OPT_SORT_KEYS |
        OPT_STRICT_INTEGER)

except ImportError:
    import json
    OPT_JSON_FLAGS = None


try:
    CompressorException = required('zstd.Error', quiet=True)
    from zstd import compress, decompress

except ImportError:
    from gzip import BadGzipFile as CompressorException
    from gzip import compress, decompress


try:
    EncoderException = required('base2048.DecodeError', quiet=True)
    import base2048

except ImportError:
    EncoderException = ValueError



class SerializeError(Exception):
    ...


def default_serializer(obj, throw=True):

    if obj is Nothing:
        return None

    if (
        isinstance(obj, tuple) and
        type(obj).__mro__ == (type(obj), tuple, object)
    ):
        return {f'<{type(obj).__name__}>': obj._asdict()}

    with suppress(KeyError):
        result = serializers[class_of(obj)](obj)
        if result is not Nothing:
            return result

    for types, callback in serializers.items():
        if isinstance(obj, types):
            result = callback(obj)
            if result is not Nothing:
                return result

    if throw:
        msg = f"couldn't serialize {Who.Is(obj)}"
        raise TypeError(msg)

    return Nothing


def serializer(*classes):

    direct_call = len(classes) == 2 and is_function(classes[1])  # noqa: PLR2004

    def name(obj):
        return f'{Who(obj, addr=True)} ({sourcefile(obj)})'

    def serialize(func):

        order = [classes[0]] if direct_call else classes
        for cls in order:

            if isinstance(cls, bytes | str):
                cls = required(cls)  # noqa: PLW2901

            if cls in serializers:
                if (
                    Who.Name(serializers[cls]) == Who.Name(func) and
                    sourcefile(serializers[cls]) == sourcefile(func)
                ):
                    continue

                logger.warning(
                    f'{Who(cls)} already have registered serializer '
                    f"{name(serializers[cls])}, can't add another "
                    f'{name(func)}', trace=True, shift=-1)

                continue
            serializers[cls] = func

        title = Who(func)
        msg = ', '.join(map(Who, order))
        if not (direct_call or title.endswith('.<lambda>')):
            msg = f'{msg} -> {title}'

        logger.verbose(msg)
        return func

    return serialize(classes[1]) if direct_call else serialize


@Monkey.wrap(json, 'dumps')
def to_json(func, data, /, **kw):

    minify = kw.pop('minify', True)
    option = kw.pop('option', Nothing)

    if OPT_JSON_FLAGS:
        # enabled only when orjson used and imported
        flags = (option or (0x0 if minify else OPT_INDENT_2))
        kw['option'] = OPT_JSON_FLAGS | flags

    elif option is not Nothing:
        # intercept orjson option when orjson not available
        logger.warning(
            f'{option=} passed, but is not supported by '
            f'stdlib json, install orjson', trace=True)

    elif not minify:
        # stdlib json with indent option
        kw.setdefault('indent', 2)
        kw.setdefault('sort_keys', True)

    kw.setdefault('default', default_serializer)
    encode = (to_ascii, to_bytes)[bool(kw.pop('bytes', False))]
    return encode(func(data, **kw))


if OPT_JSON_FLAGS:
    json.JSONDecodeError = JSONDecodeError


@Monkey.bind(json, 'repr')
def try_json(data, /, **kw):
    with suppress(Exception):
        if (
            hasattr(data, 'as_dict') and
            (as_dict := data.as_dict) and
            isinstance(as_dict, dict)
        ):
            data = dict(as_dict)

    try:
        return to_json(data, **kw)

    except Exception:  # noqa: BLE001
        return repr(data)


@Monkey.bind(json)
def cast(obj):

    if is_mapping(obj):
        return {k: cast(v) for k, v in obj.items()}

    elif isinstance(obj, tuple) and hasattr(obj, '_fields'):  # namedtuple
        return cast(obj._asdict())

    elif isinstance(obj, deque | list | tuple | set):
        return list(map(cast, obj))

    elif is_collection(obj):
        msg = f"couldn't serialize {Who.Is(obj)}"
        raise TypeError(msg)

    result = default_serializer(obj, throw=False)
    return obj if result is Nothing else result


class Encoding:
    Base16   = base64.b16encode, base64.b16decode, 'ascii'
    Base32   = base64.b32encode, base64.b32decode, 'ascii'
    Base64   = base64.b64encode, base64.b64decode, 'ascii'
    Base85   = base64.b85encode, base64.b85decode, 'ascii'
    Codecs = [Base16, Base32, Base64, Base85]  # noqa: RUF012

    if EncoderException is not ValueError:
        Base2048 = base2048.encode, base2048.decode, 'utf-8'
        Codecs.append(Base2048)

    Last = Codecs[-1]
    Codecs = tuple(Codecs)

    Charsets: ClassVar[dict[str, tuple]] = {
        '0123456789ABCDEF': Base16,
        '234567=ABCDEFGHIJKLMNOPQRSTUVWXYZ': Base32,
        '+/0123456789=ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz': Base64,
        (
            '!#$%&()*+-0123456789;<=>?@ABCDEFGHIJKLMNOPQ'
            'RSTUVWXYZ^_`abcdefghijklmnopqrstuvwxyz{|}~'
        ): Base85,
    }


def dumps(value, /, encoder=None, proto=None, ratio=None, charset=None):
    """Pickle & compress any object with any pickler with .dumps/loads methods
    and store module name with value for automatic unpickling."""

    if encoder is None:
        value = generic_dumps(value, proto or -1)

    else:
        encoder = to_bytes(encoder)
        if len(encoder) >= 32 - len(DEADBEEF):
            msg = (
                f'len({encoder=})={len(encoder):d} '
                f'must be lower than {32 - len(DEADBEEF):d}')
            raise ValueError(msg)

        module = required(BACKENDS.get(encoder) or encoder)
        value = module.dumps(value)
        value = encoder + DEADBEEF + to_bytes(value, charset=charset or 'utf-8')

    return compress(value, ratio or 9)


def loads(value, /, expect=None):
    """Decompress pickled by dumps function objects with autodetect used module
    for marshallization.

    by default, expecting four exceptions for back compatibility with
    default dumb cache and pass all errors from cache
    """

    if value is None:
        return

    if expect is None:
        expect = TypeError, ValueError, UnpicklingError, CompressorException

    try:
        value = decompress(value)

    except expect:
        with suppress(Exception):
            return json.loads(value)
        return

    offset = value[:32].find(DEADBEEF)
    if offset != -1:
        encoder = to_bytes(value[:offset])
        return (
            required(BACKENDS.get(encoder) or encoder)
            .loads(value[offset + len(DEADBEEF):]))

    with suppress(expect):
        return generic_loads(value)  # noqa: S301


def pack(value, *args, codec=Encoding.Last, **kw):
    """Pack value with codec prefix and encode it with codec"""

    return (
        bytes(hex(Encoding.Codecs.index(codec)), 'ascii')[2:] +
        Str.to_bytes(codec[0](to_bytes(dumps(value, *args, **kw))), codec[2])
    ).decode(codec[2])


def unpack(string, *args, **kw):
    """Unpack string with codec prefix and try to decode it with codec"""

    last = kw.pop('last', False)
    try:
        codec = Encoding.Codecs[int(string[0])][1]
        return loads(codec(string[1:]), *args, **kw)

    except (IndexError, ValueError, Error):
        if last:
            return

        with suppress(EncoderException, ValueError, Error):
            for charset in reversed(Encoding.Charsets):
                codec = Encoding.Codecs.index(Encoding.Charsets[charset])
                result = unpack(string.__class__(codec) + string, last=True)
                if result:
                    return result

            codec = Encoding.Codecs.index(Encoding.Last)
            result = unpack(string.__class__(codec) + string, last=True)
            if result:
                return result

        return loads(string, *args, **kw)


RESERVED_KEYSET = frozenset(kwlist)
@cache(2 * 10)
def namedtuple_builder(*fields, name=None):

    fields_set = set(fields)
    if len(fields) != len(fields_set):
        raise ValueError(f'{fields=} must be unique')

    if RESERVED_KEYSET & fields_set:
        reserved = tuple(sort(RESERVED_KEYSET & fields_set))
        raise ValueError(f"{fields=} must not contain keywords what's {reserved=}")

    return namedtuple(name or 'Tuple', fields)


def to_namedtuple(*args, **kw):
    name = kw.pop('name', None)
    rename = kw.pop('rename', False)
    sort_keys = kw.pop('sort_keys', True)

    if len(args) == 1 and not kw:
        return make_namedtuple(args[0], name, rename, sort_keys)

    elif (
        not args and kw  # this is to_namedtuple(**data) style
    ):
        if name is None and rename is False and sort_keys is True:
            return make_namedtuple(kw, name, rename, sort_keys)

        fields = []
        if name is not None:
            fields.append(name)
        if rename is not False:
            fields.append(rename)
        if sort_keys is not True:
            fields.append(sort_keys)
        fields.sort()

        raise TypeError(
            'to_namedtuple() used with **kw style but kw contains '
            f'to_namedtuple{fields} reserved keys')

    raise TypeError(f'to_namedtuple({args=}, {kw=})')


def make_namedtuple(data, name, rename, sort_keys):

    if isinstance(data, tuple) and hasattr(data, '_asdict'):
        return data

    elif isinstance(data, optional('immutabledict.immutabledict', default=dict)):
        ...

    elif not isinstance(data, dict):
        raise TypeError(f'expected dict, got {Who.Is(data)}')

    keys = tuple(sort(data) if sort_keys else data)

    if not rename:
        fields = keys[:]
    else:
        fields = tuple(f'{k}_' if k in RESERVED_KEYSET else k for k in keys)

    return namedtuple_builder(*fields, name=name)(*map(data.__getitem__, keys))


with suppress(ImportError):
    from pydantic import BaseModel

    @serializer(BaseModel)
    def from_pydantic(obj):
        return obj.model_dump()



from collections import defaultdict
from contextlib import suppress
from enum import EnumMeta, IntEnum
from functools import wraps
from http import HTTPStatus
from http.cookies import SimpleCookie
from inspect import iscoroutine
from pathlib import Path
from re import split, sub
from typing import ClassVar

from kalib.dataclass import dataclass
from kalib.datastructures import Encoding, json, loads, pack, serializer, unpack
from kalib.descriptors import Property, cache, parent_call
from kalib.importer import optional, required, sort
from kalib.internals import Who
from kalib.loggers import Logging
from kalib.misc import lazy_proxy_to
from kalib.text import Str


def build_enumerate(name, *order):
    if not isinstance(name, str):
        msg = (
            f'first arg, class name of build_enumerate() '
            f'must be string, not {Who(name)}')
        raise TypeError(msg)

    types = tuple(set(map(type, order)))


    if len(types) != 1:
        msg = f"isn't implemented mixed attributes order: {types!r}"
        raise ValueError(msg)

    if len(order) == 1 and types[0] is dict:
        order = order[0]

    elif types[0] is tuple:

        def titler(code, phrase):
            order = phrase.split(' ')
            if code // 100 >= 6:  # noqa: PLR2004
                order = [name.upper(), *order]
            return '_'.join(map(str.upper, order))

        order = {
            titler(code, phrase):
            (code, phrase, description)
            for code, phrase, description in order}

    elif issubclass(types[0], EnumMeta):
        result = {}
        for klass in order:
            for i in klass:
                value = getattr(klass, i.name)
                if i.name in order:
                    raise ValueError

                result[i.name] = value.value, value.phrase, value.description
        order = result

    else:
        msg = f"make enum from {types[0]!r} isn't implemented"
        raise ValueError(msg)

    class Special(dict):
        @Property.Cached
        def _member_names(self):
            return tuple(filter(lambda x: not x.startswith('_'), self))

    def __new__(cls, value, phrase, description=''):  # noqa: N807
        obj = int.__new__(cls, value)
        obj._value_ = value

        obj.phrase = phrase
        obj.description = description
        return obj

    order['__new__'] = __new__

    return type(name, (IntEnum,), Special(order))


HTTPStatus = build_enumerate(
    'HTTPStatus', {
        i.name: (i.value, i.phrase) + ((i.description,)
        if i.description else ()) for i in HTTPStatus})

CloudFlare = build_enumerate(
    'CloudFlare',
    (520, 'Cloudflare Unknown Error', 'Web server returns an Unknown Error'),
    (521, 'Cloudflare Server Down', 'Web server is down'),
    (522, 'Cloudflare Connection Timeout', 'Connection timed out'),
    (523, 'Cloudflare Unreachable', 'Origin is unreachable'),
    (524, 'Cloudflare Timeout', 'A timeout occurred'),
    (525, 'Cloudflare Handshake Failed', 'SSL handshake failed'),
    (526, 'Cloudflare Invalid Certificate', 'Invalid SSL certificate'),
    (527, 'Cloudflare Railgun Error', 'Railgun Listener to origin error'))

Database = build_enumerate(
    'Database',
    (600, 'General Failure', 'Common error'),
    (601, 'Bad Scheme', 'Broken or incompatible scheme'),
)

Crawling = build_enumerate(
    'Crawling',
    (700, 'General Failure', 'Common error'),
    (701, 'Mitigation Hit', 'Cloudflare challenge mitigation received'),
    (702, 'Temporary Unavailable', 'Resource temporary unavailable'),
    (703, 'Still Processing', 'Resource temporary unavailable: video compressing, etc'),
    (704, 'Need Reparse', 'Resource unavailable: need to reparse'),
    (705, 'Restricted', 'Resource unavailable: under age banner'),
    (706, 'Missing Data', 'Resource unavailable: missing fields, migration issues, etc'),
    (707, 'Need Subscription', 'Resource unavailable: paywall, need account, etc'),
    (708, 'Regional Restrictions', 'Resource unavailable for current region'),
    (709, 'No Valid Source', 'Resource unavailable: no valid video format'),
    (710, 'Broken Data', 'Resource unavailable: bad bitstream, failed transfer, etc'),
    (711, 'Missing Required Data',
          'Resource unavailable: missing required data, e.g. duration'),
)

HTTP = build_enumerate(
    'HTTP', HTTPStatus, CloudFlare, Database, Crawling)


class MimeType(dataclass):

    type : str | None
    text : str | None


class ResponseInternals(dataclass):

    url: str | \
        optional('yarl.URL', default=str) | \
        optional('httpx.URL', default=str)

    status  : int
    reason  : str | None  # can be None for disk cache
    headers : dict[str, str]
    content : bytes | Path  # can be path for disk cache


def bind_main_classes_as_property(cls):
    for code, exception in cls.exceptions.items():
        if code < 100:  # noqa: PLR2004
            setattr(cls, cls.RFC9110[code], exception)
    return cls


class HTTPResponse(Logging.Mixin):

    @classmethod
    def read(cls, response, /, **kw):

        if (
            isinstance(response, dict) and
            isinstance(response.get('headers'), Path) and
            isinstance(response.get('content'), Path)
        ):
            return FileResponse._read(response, response.get('url'), **kw)  # noqa: SLF001

        if iscoroutine(response):
            msg = f"can't process {response=}, it's coroutine, await it first"
            raise RuntimeError(msg)

        order = {
            HTTPxResponse    : 'httpx.Response',
            RequestsResponse : 'requests.models.Response',
            AioHttpResponse  : 'aiohttp.client_reqrep.ClientResponse'}

        for backend, path in order.items():
            with suppress(ImportError):
                if isinstance(response, required(path, quiet=True)):
                    return backend._read(response, **kw)  # noqa: SLF001

        raise NotImplementedError(f"{Who.Is(response)} isn't implemented")

    # common interface

    def __init__(self, response, content):
        self._raw = response
        self._content = content

    @Property.Cached
    def _headers(self):
        result = defaultdict(list)
        for key, value in self._raw.headers.items():
            result[key].append(value)
        return {
            k: tuple(v) if len(v) > 1 else v[0] for k, v in result.items()}

    @Property.Cached
    def _response_params(self):
        return {
            'url'     : self._raw.url,
            'reason'  : self._raw.reason,
            'headers' : self._headers,
            'content' : self._content}

    @Property.Cached
    def _response(self):
        return ResponseInternals.load(self._response_params)

    @Property.Cached
    def ok(self):
        return not self.exception.not_ok

    # just mapping to common internal structure

    @Property.Cached
    def url(self):
        return str(self._response.url)

    @Property.Cached
    def status(self):
        return int(self._response.status)

    @Property.Cached
    def reason(self):
        if (reason := self._response.reason):
            return reason

        reason = HTTPException.Statuses[self.status]
        return f'{reason.phrase}: {reason.description}'

    @Property.Cached
    def headers(self):
        return self._response.headers

    @Property.Cached
    def headerstring(self):
        return json.repr(self.headers)

    @Property.Cached
    def content(self):
        return self._response.content

    @Property.Cached
    def exception(self):
        return HTTPException.by_code(self.status)

    @Property.Cached
    def mime(self):
        msg = (
            f"could't detect mime-type for {self.url!r} response: "
            f'{self.headerstring}: {self.content!r}')

        if not self.content:
            self.log.verbose(msg)
            return MimeType.load({})

        try:
            if (mime := Str(self.content).mime):
                result = MimeType.load(mime)
            else:
                self.log.verbose(msg)
                return MimeType.load({})

        except Exception:
            self.log.exception(msg)
            raise

        self.log.debug(
            f"mime-type for {self.url!r} is {result.type!r}, it's {result.text!r} ")
        return result

    # content related properties

    @Property.Cached
    def bytes(self):
        return Str.to_bytes(self.content)

    @Property.Cached
    def text(self):
        return Str.to_str(self.content)

    @Property.Cached
    def feed(self):
        return required('feedparser.parse')(self.content)

    @Property.Cached
    def json(self):
        return json.loads(self.content)

    @Property.Cached
    def html(self):
        return required('lxml.html.document_fromstring')(self.content)

    @Property.Cached
    def pack(self):
        return required('msgpack.loads')(self.content, encoding='utf-8', use_list=False)

    @Property.Cached
    def xml(self):
        return required('lxml.etree').fromstring(self.content)

    # header-based mapping

    decoders = {  # noqa: RUF012

        # feedparser
        'application/atom+xml'  : 'feed',
        'application/rdf+xml'   : 'feed',
        'application/x-rss+xml' : 'feed',
        'text/x-opml'           : 'feed',
        'application/rss+xml'   : 'feed',

        # lxml html
        'application/xhtml+xml' : 'html',
        'text/html'             : 'html',

        # lxml xml
        'application/xml'       : 'xml',
        'application/rsd+xml'   : 'xml',
        'text/xml'              : 'xml',
        'xml'                   : 'xml',

        # other default types
        'application/json'      : 'json',
        'application/x-msgpack' : 'pack',
        'text/plain'            : 'text',
    }

    @Property.Cached
    def content_type(self):
        def getter(x):
            with suppress(KeyError):
                return self.headers.get(x, '').lower().split(';', 1)[0].strip()

        for header in ('Content-Type', 'content-type'):
            if (content_type := getter(header)):
                return content_type

    @Property.Cached
    def data(self):

        @cache
        def get_message(reason):
            return f'{reason} for {self.url!r} response ({self.headerstring})'

        try:
            content_type = self.content_type
            method = self.decoders[content_type]

        except KeyError:
            message = get_message("couldn't select callback")
            self.log.warning(f'try {content_type=}, {message}')

            try:
                content_type = mime_type = self.mime.type
                method = self.decoders[mime_type]
            except Exception:
                self.log.exception(f'try {mime_type=}, {message}')
                raise

        try:
            return getattr(self, method)

        except Exception:
            message = get_message(
                f'something went wrong on {method=}')
            self.log.exception(f'by {content_type=}, {message}')
            raise


class AioHttpResponse(HTTPResponse):

    @classmethod
    async def _read(cls, response, /, **kw):
        return HTTPException.catch(
            cls(response, content=await response.read()), **kw)

    @Property.Cached
    @parent_call
    def _response_params(self, parent):
        return parent | {'status': self._raw.status}


class RequestsResponse(HTTPResponse):

    @classmethod
    def _read(cls, response, /, **kw):
        return HTTPException.catch(
            cls(response, content=response.content), **kw)

    @Property.Cached
    @parent_call
    def _response_params(self, parent):
        return parent | {'status': self._raw.status_code}


class HTTPxResponse(HTTPResponse):

    @classmethod
    def _read(cls, response, /, **kw):
        return HTTPException.catch(
            cls(response, content=response.content), **kw)

    @Property.Cached
    def _response_params(self):
        return {
            'url'     : self._raw.url,
            'content' : self._content,
            'headers' : self._headers,
            'reason'  : self._raw.reason_phrase,
            'status'  : self._raw.status_code,
        }


class FileResponse(HTTPResponse):

    @classmethod
    def _read(cls, data, url=None, /, **kw):
        with data['headers'].open('rb') as fd:
            headers = json.loads(fd.read())

        self = cls(headers, content=data['content'])
        if url := (headers.get('url') or url):
            headers['url'] = url
        else:
            raise ValueError(f"can't get url from {headers=}")
        return HTTPException.catch(self, **kw)

    @Property.Cached
    def _headers(self):
        headers = dict(self._raw)
        del headers['status']
        del headers['url']
        return headers

    @Property.Cached
    def _response_params(self):
        headers = dict(self._raw)
        return {
            'url'     : headers['url'],
            'status'  : headers['status'],
            'headers' : self._headers,
            'content' : self._content,
            'reason'  : None,
        }

    @Property.Cached
    def content(self):
        with self._content.open('rb') as fd:
            return fd.read()


@lazy_proxy_to('state', 'description', 'phrase', 'value', Property.Class.Cached)
@bind_main_classes_as_property
class HTTPException(Exception, Logging.Mixin):  # noqa: N818

    RFC9110: ClassVar[dict] = {
        1: 'Information',
        2: 'Successful',
        3: 'Redirection',
        4: 'Client',
        5: 'Server',
        6: 'Database',
        7: 'Crawling',
    }
    Statuses: ClassVar[dict] = {int(x.value): x for x in HTTP}

    @staticmethod
    def scream_to_snake(text):
        return sub(
            r'((^[a-z])|(_[a-z]))',
            lambda x: x.group(1)[-1].upper(), text.lower())

    @Property.Class.Parent
    def exceptions(cls):
        @cache
        def make_parent_class(no):
            prefix = 'HTTP' if no <= 5 else ''  # noqa: PLR2004
            return type(f'{prefix}{cls.RFC9110[no]}Error', (cls,), {})

        result = {}
        classes = [[], []]

        for status in HTTP:
            offset = status.value // 100
            section = cls.RFC9110[offset]
            name = sub(r'(Error)$', '', cls.scream_to_snake(status.name))

            parent = make_parent_class(offset)
            if offset not in result:
                result[offset] = parent
                classes[bool(offset >= 3)].append(parent)  # noqa: PLR2004
                setattr(HTTP, section, parent)

            root = HTTP if offset in (2, 3) else parent
            if offset > 5:  # noqa: PLR2004
                prefix = ''
            elif offset in (2, 3):
                prefix = 'HTTP'
            else:
                prefix = f'HTTP{section}'

            postfix = 'Error' if offset >= 4 else ''  # noqa: PLR2004
            result[status.value] = type(
                f'{prefix}{name}{postfix}', (parent,), {'state': status})

            short = name.removeprefix(section)
            setattr(root, short, result[status.value])

        cls.Allright = tuple(classes[0])
        cls.Exception = tuple(classes[1])

        return result

    @Property.Class.Cached
    def not_ok(cls):
        return not issubclass(cls, cls.Allright)

    @classmethod
    def by_code(cls, *status, throw=False):
        if len(status) > 1:
            return tuple(map(cls.by_code, status))
        status = status[0]

        if not isinstance(status, HTTP | HTTPException | int):
            msg = (
                f'status code should be {Who(HTTP)} | '
                f'{Who(HTTPException)} | int, '
                f'but not {Who(status)}')
            raise TypeError(msg)

        if isinstance(status, HTTPException):
            status = status.state.value

        if isinstance(status, int):
            try:
                status = cls.Statuses[status]
            except KeyError:
                cls.log.warning(
                    f'unknown status code in response: '
                    f"couldn't get {Who(cls)}({status})")
                if not throw:
                    return HTTPUnknownStatusError
                raise

        return cls.exceptions[int(status.value)]

    def __init__(self, response):
        self.response = response

    @classmethod
    def catch(cls, response, include=None, exclude=None):
        if iscoroutine(response):
            msg = f"can't process {response=}, it's coroutine, await it first"
            raise RuntimeError(msg)

        def to_int(x):
            if isinstance(x, HTTPException):
                return int(x.value)

            elif isinstance(x, int):
                return x

            raise ValueError(
                f'include/exclude can only contains int | {Who(HTTPException)}, '
                f'not {Who.Is(x)} which passed')

        # always raise included exceptions

        if (include and (
            isinstance(include, int) or issubclass(include, HTTPException)
        )):
            include = (include,)
        include = tuple(map(to_int, include or ()))

        # do not raise excluded exceptions

        if (exclude and (
            isinstance(exclude, int) or issubclass(exclude, HTTPException)
        )):
            exclude = (exclude,)
        exclude = tuple(map(to_int, exclude or ()))

        code = int(response.status)
        if code in (exclude or ()):
            return response

        elif code // 100 >= 4 or code in (include or ()):  # noqa: PLR2004, raise all >=400
            raise cls.by_code(code)(response)

        return response

    @Property.Cached
    def args(self):
        return (
            self.response.url,
            self.response.status,
            self.response.reason,
            self.response.headerstring,
            self.response.mime.as_dict if self.response.mime else {},
        )

    @Property.Cached
    def verbose(self):
        response = self.response

        try:
            body = f'{json.repr(response.json)=}'
        except Exception:  # noqa: BLE001
            body = response._response.content  # noqa: SLF001

        try:
            return (
                f'{response.url=}\n'
                f'{json.repr(response.headers)=}\n{body=}')

        except Exception:
            self.log.exception(f'something went wrong with {response=}')
            raise

    def __str__(self):
        return f'{self.response.status:d} {self.response.reason}'


class HTTPUnknownStatusError(HTTPException):
    ...


@lazy_proxy_to('generator', 'random')
class Agent:
    def __init__(self, browsers=None, min_percentage=None, os=None):
        self._kw = {
            'os'       : list(os or ('windows', 'macos', 'linux')),
            'browsers' : list(browsers or ('edge', 'chrome')),
            'min_percentage': min_percentage or 1.0}

    @Property.Cached
    def generator(self):
        return required('fake_useragent.UserAgent')(**self._kw)

    @property
    def any(self):
        return self.generator.random

    @Property.Class
    def header(cls):
        return {'User-Agent': cls().any}


@lazy_proxy_to('cookies', 'keys', 'values', '__iter__')
class Cookies:
    skip_keys = ('domain', 'expires', 'path')

    @classmethod
    def from_dump(cls, data):
        if data is None:
            return

        elif not isinstance(data, bytes | str):
            msg = f'accept only bytes | str as data, not {Who.Is(data)}'
            cls.log.fatal(msg)
            raise TypeError(msg)

        decoded = unpack(data) or loads(data)
        if not decoded:
            msg = f"couldn't interpret received {data=}"
            cls.log.fatal(msg)
            raise ValueError(msg)

        return cls(decoded)

    @classmethod
    def from_env(cls, key):
        if key is None:
            return

        elif not isinstance(key, bytes | str):
            msg = (
                f'accept only bytes | str as environment '
                f'variable key, not {Who.Is(key)}')
            cls.log.fatal(msg)
            raise TypeError(msg)

        data = required('environment.env').str(key, None)
        if not data:
            cls.log.verbose(f"failed, {key=} isn't exists or empty", once=True)
            return

        self = cls.from_dump(data)
        self.log.verbose(f'{key=} got {self.as_json}', once=True)
        return self


    @classmethod
    def from_file(cls, path):
        if path is None:
            return

        elif not isinstance(path, bytes | str):
            msg = (
                f'accept only bytes | str as filesystem '
                f'path, not {Who.Is(path)}')
            cls.log.fatal(msg)
            raise TypeError(msg)

        path = Path(path)
        if not path.is_file():
            cls.log.verbose(f"failed, {path=} isn't exists or accessible")
            return

        with path.open('r+b') as fd:
            self = cls(fd.read())

        path = str(path)
        self.log.verbose(f'{path=} got {self.as_json}', once=True)
        return self

    def __init__(self, raw):
        self._raw = raw

    def __bool__(self):
        return bool(self.as_data)

    def __repr__(self):
        return f'<{self.selfname}: {self.as_json}>'

    @classmethod
    def iterstrings(cls, iterable):
        yield from filter(bool, map(str.strip, iterable))

    def iterraw(self):
        raw = self._raw

        if isinstance(raw, bytes | str):
            lines = Str.to_ascii(raw).split('\n')

        elif isinstance(raw, SimpleCookie):
            lines = [i.OutputString() for i in raw.values()]

        elif isinstance(raw, optional('aiohttp.CookieJar')):
            lines = [i.OutputString() for i in raw]

        else:
            lines = list(raw)
            self.log.debug(f'try to iterate over {Who.Is(lines)}')

        yield from self.iterstrings(lines)

    @Property.Cached
    def cookies(self):
        def iter_tokens():
            for line in self.iterraw():
                yield from self.iterstrings(line.split(';'))

        def iter_values():
            for line in iter_tokens():
                match = split(r'([\s=]+)', line, maxsplit=1)
                if len(match) == 3 and match[0] not in self.skip_keys:  # noqa: PLR2004
                    yield f'{match[0]}={match[2]}'

        cookie = None
        result = SimpleCookie()

        for line in sort(iter_values()):
            cookie = f'{cookie}; {line}' if cookie else line
            if len(cookie) >= 2 ** 10:
                result.load(cookie)
                cookie = None

        if cookie:
            result.load(cookie)

        return result

    @Property.Cached
    def as_data(self):
        return tuple(sort(i.OutputString() for i in self.cookies.values()))

    @Property.Cached
    def as_dict(self):
        return {i.key: i.value for i in self.values()}

    @Property.Cached
    def as_json(self):
        return json.dumps(self.as_data)

    @Property.Cached
    def as_text(self):
        return '\n'.join(self.as_data)

    @Property.Cached
    def as_base(self):
        return pack(self.as_data, codec=Encoding.Base85, encoder='json')[1:]


def proxy_to_super(func):
    @wraps(func)
    def wrapper(self, other):
        return self.load(func(self)(other).as_dict)
    return wrapper


class URL(dataclass):
    scheme   : str
    host     : str | int
    port     : str | int | None = None
    user     : str | None = None
    password : str | None = None
    path     : str | None = '/'
    query    : str = ''
    fragment : str = ''

    @Property.Class.Cached
    def subclass(cls):
        return required('yarl.URL')

    @Property.Class.Cached
    def default_ports(cls):
        return {
            'http'  : 80,
            'ftp'   : 21,
            'ssh'   : 22,
            'https' : 443,
            'amqp'  : 5672,
            'redis' : 6379,
        }

    @classmethod
    def clean_kwargs(cls, config):
        port = config.get('port')

        with suppress(KeyError):
            if config['path'] == '/':
                del config['path']

        if (
            port and
            (isinstance(port, int) or (isinstance(port, str) and port.isdigit())) and
            (port := int(port)) >= 0
        ):
            with suppress(KeyError):
                if cls.default_ports[config['scheme'].lower()] == port:
                    del config['port']

        return {k: v for k, v in config.items() if v}

    @classmethod
    def _preload(cls, config):
        if isinstance(config, str):
            config = cls.subclass(config)

        elif isinstance(config, dict | dict):
            return cls.clean_kwargs(dict(config))

        with suppress(KeyError):
            userattr = {
                'furl.furl.furl' : 'username',
                'yarl.URL'       : 'user',
            }[Who(config)]

            port = config.port or None
            config = {
                'scheme'  : str(config.scheme).lower(),
                'user'    : getattr(config, userattr),
                'password': config.password,
                'host'    : config.host,
                'port'    : port,
                'path'    : str(config.path or ''),
                'query'   : config.query_string or '',
                'fragment': config.fragment or ''}

            return cls.clean_kwargs(config)

        raise TypeError(f'{Who(cls)} unknown input: {Who.Is(config)}')

    @Property.Cached
    def as_dict(self):
        return {
            'scheme'   : self.scheme,
            'user'     : self.user,
            'password' : self.password,
            'host'     : self.host,
            'port'     : self.port,
            'path'     : self.path,
            'query'    : self.query,
            'fragment' : self.fragment}

    @Property.Cached
    def url(self):
        return self.subclass.build(**self.as_dict)

    def __str__(self):
        return str(self.url)

    @proxy_to_super
    def __and__(self):  # noqa: PLE0302
        return super().__and__

    @proxy_to_super
    def __xor__(self):  # noqa: PLE0302
        return super().__xor__

    @proxy_to_super
    def __sub__(self):  # noqa: PLE0302
        return super().__sub__

    @proxy_to_super
    def __or__(self):  # noqa: PLE0302
        return super().__or__

    @proxy_to_super
    def __add__(self):  # noqa: PLE0302
        return super().__add__

# make all exceptions for easy imports at start time

for exception in HTTPException.exceptions.values():
    locals()[exception.__name__] = exception

@serializer(HTTPException)
def from_http(something):
    return something.state.value

HTTP.URL = URL
HTTP.Agent = Agent
HTTP.Catch = HTTPResponse.read
HTTP.Exception = HTTPException



from collections import deque, namedtuple
from collections.abc import Callable, Collection, Iterable, Mapping, Sequence
from contextlib import suppress
from functools import cache, partial
from inspect import (
    getmodule,
    getsourcefile,
    isawaitable,
    isbuiltin,
    isclass,
    iscoroutine,
    isfunction,
    ismethod,
    ismodule,
    stack,
)
from itertools import filterfalse
from operator import itemgetter
from os.path import splitext
from pathlib import Path
from platform import architecture
from re import search, sub
from sys import modules
from sysconfig import get_paths
from traceback import extract_stack, format_stack
from types import FunctionType, LambdaType, UnionType
from typing import Any, Generic, Union, get_args, get_origin

Collections = deque, dict, list, set, tuple, bytearray
Primitives  = bool, float, int, str, complex, bytes
Builtins    = Primitives + Collections

is_windows = 'windows' in architecture()[1].lower()


def is_callable(obj):
    return isinstance(obj, Callable) or callable(obj)


def is_collection(obj):
    return (
        isinstance(obj, Collection) and
        isinstance(obj, Sequence) and
        not isinstance(obj, bytes | str)
    ) or issubclass(class_of(obj), Collections)


def is_mapping(obj):
    if isinstance(obj, Mapping):
        return True

    cls = class_of(obj)
    if issubclass(cls, dict):
        return True

    for key in ('__getitem__', '__setitem__', '__delitem__'):
        if not hasattr(obj, key):
            return False

    return True


def is_iterable(obj):
    return isinstance(obj, Iterable) or hasattr(obj, '__iter__')


def is_from_primivite(obj):
    """
    Check if the object instantiated from primitive type.
    """
    return bool(obj is None or isinstance(obj, Primitives))


def is_from_builtin(obj):
    return bool(isinstance(obj, Collections) or is_from_primivite(obj))


def is_primitive(obj):
    return (
        obj is True or obj is False or
        obj is None or type(obj) in Builtins)


def unique(iterable, /, key=None, include=None, exclude=None):
    skip = include is None

    if not key:
        exclude = set(exclude or ())
        include = frozenset(include or ())

    else:
        exclude = set(map(key, exclude or ()))
        include = frozenset(map(key, include or ()))

    excluded = exclude.__contains__
    included = include.__contains__
    is_dict = isinstance(iterable, dict)

    for element in iterable:

        k = key(element) if key else element
        if not excluded(k) and (skip or included(k)):

            yield (element, iterable[element]) if is_dict else element
            exclude.add(k)

@cache
def trim_module_path(full):
    dirs = get_paths()

    path = str(full)
    if is_windows:
        path = path.lower()

    for scheme, reason in (
        ('stdlib', True),
        ('purelib', False),
        ('platlib', False),
        ('platstdlib', True),
    ):
        subdir = dirs[scheme]
        if is_windows:
            subdir = subdir.lower()

        if path.startswith(subdir):
            return reason, str(full)[len(subdir) +1:]

    subdir = str(Path(__file__).parent.parent)

    if is_windows:
        subdir = subdir.lower()

    if path.startswith(subdir):
        return False, str(full)[len(subdir) +1:]

    return None, str(full)


def is_internal(x):
    if (isbuiltin(x) or isbuiltin(class_of(x))):
        return True

    elif (module := get_module(x)):

        if module.__name__ == 'builtins':
            return True

        is_stdlib = trim_module_path(module.__file__)[0]
        if is_stdlib is not None:
            return is_stdlib

    return False


def class_of(obj):
    return obj if isclass(obj) else type(obj)


def issubstance(obj, types):  # noqa: PLR0911, PLR0912
    if types is None:
        return False

    if types in (Any, obj):
        return True

    if get_origin(types) in (Generic, Union, UnionType):
        klass = class_of(obj)
        included_types = get_args(types)
        if Any in included_types:
            return True

        for i in included_types:
            try:
                if issubstance(klass, i):
                    return True
            except TypeError:
                continue
        return False

    if is_collection(types) and not isinstance(types, tuple):
        types = tuple(types)

    if is_collection(types):
        if not types:
            return False

        elif Any in types:
            return True

    if get_args(types):
        types = get_origin(types)

    if types in (Any, obj):
        return True

    if is_collection(types) and obj in types:
        return True

    try:
        if (result := (
            isclass(obj) and
            (obj is types or issubclass(obj, types)))
        ):
            return result

        if not is_collection(types):
            return isinstance(obj, types)

        for sometype in types:
            if (result := issubstance(obj, sometype)):
                return result

    except TypeError as e:
        msg = f'invalid check what {Who.Is(obj)} is instance of {Who.Is(types)}'
        raise TypeError(msg) from e


def iter_inheritance(  # noqa: PLR0913
    obj,
    include        = None,
    exclude        = None,
    exclude_self   = True,
    exclude_stdlib = True,
    reverse        = False,
):
    order = class_of(obj).__mro__[:-1]

    if not exclude_self:
        order = unique((obj, *order), key=id)
    else:
        order = unique(filter(lambda x: x is not obj, order), key=id)

    if reverse:
        order = reversed(list(order))

    if include:
        if isinstance(include, FunctionType | LambdaType):
            order = filter(include, order)
        else:
            if not is_iterable(include):
                include = (include,)
            order = filter(include.__contains__, order)

    if exclude:
        if isinstance(exclude, FunctionType | LambdaType):
            order = filterfalse(exclude, order)
        else:
            if not is_iterable(exclude):
                exclude = (exclude,)
            order = filterfalse(exclude.__contains__, order)

    if exclude_stdlib:
        order = filterfalse(is_internal, order)

    yield from order


def get_mro(obj, /, **kw):

    func = kw.pop('func', None)
    glue = kw.pop('glue', None)

    result = iter_inheritance(obj, **kw)

    if func:
        result = tuple(map(func, result))

    if glue:
        result = glue.join(result)

    return result if (func or glue) else tuple(result)


def get_entity(something, name, **kw):

    index = kw.pop('index', 0)
    kw.setdefault('exclude_self', False)
    kw.setdefault('exclude_stdlib', False)

    counter = 0
    for obj in iter_inheritance(something, **kw):
        try:
            attr = obj.__dict__[name]
            if not counter - index:
                return attr, obj
            counter += 1
        except KeyError:
            continue
    raise KeyError


def get_owner(obj, name, **kw):
    try:
        return get_entity(obj, name, **kw)[1]
    except KeyError:
        ...


def get_attr(obj, name, default=None, **kw):
    try:
        return get_entity(obj, name, **kw)[0]
    except KeyError:
        return default


def get_module(x):
    if ismodule(x):
        return x

    elif (
        (module := getmodule(x)) or
        (module := getmodule(class_of(x)))
    ):
        return module


def get_module_name(x):
    if module := get_module(x):
        with suppress(AttributeError):
            return module.__spec__.name


# This function is used to get the name of the object

def objectname(obj, full=True):
    """
    Returns the name of the given object, optionally including the full module path.

    Args:
        obj: The object whose name is to be determined.
        full (bool): If True, the full module path is included in the name.

    Returns:
        str: The name of the object, optionally including the full module path.

    Notes:
        - If the object is a module, its name is returned.
        - If the object is a coroutine, function, or method, its name is constructed
          using its module and qualified name.
        - If the object is a property, the name of its getter function is returned.
        - If the object is of any other type, the name of its class is returned.
    """

    def post(x):
        return sub(r'^([\?\.]+)', '', sub('^(__main__|__builtin__|builtins)', '', x))

    def get_module_from(x):
        return getattr(x, '__module__', get_module_name(x)) or '?'

    def get_object_name(x):
        if obj is Any:
            return 'typing.Any' if full else 'Any'

        name = getattr(x, '__qualname__', x.__name__)
        module = get_module_from(x)

        if not name.startswith(module):
            name = f'{module}.{name}'
        return name

    def main(obj):
        if ismodule(obj):
            return get_module_name(obj)

        for itis in iscoroutine, isfunction, ismethod:
            if itis(obj):
                name = get_object_name(obj)
                with suppress(AttributeError):
                    name = f'{objectname(obj.im_self or obj.im_class)}.{post(name)}'
                return name

        cls = class_of(obj)
        if cls is property:
            return get_object_name(obj.fget)

        return get_object_name(cls)

    name = post(main(obj))
    return name if full else name.rsplit('.', 1)[-1]


def Who(obj, /, full=True, addr=False): # noqa: N802
    """
    Retrieve the name of an object, optionally including its memory address.
    Parameters:
    obj (object): The object whose name is to be retrieved.
    full (bool, optional): If True, retrieve the full name of the object.
        If False, retrieve a shorter name. Defaults to True.
    addr (bool, optional): If True, append the memory address of the object
        to the name. Defaults to False.
    Returns:
    str: The name of the object, optionally including its memory address.
    """

    key = '__name_full__' if full else '__name_short__'

    def get_name():
        try:
            store = obj.__dict__
            with suppress(KeyError):
                return store[key]
        except AttributeError:
            store = None

        name = objectname(obj, full=full)
        if store is not None:
            with suppress(AttributeError, TypeError):
                setattr(obj, key, name)
        return name

    name = get_name()
    if addr:
        name = f'{name}#{id(obj):x}'
    return name


def Is(obj, /, **kw): # noqa: N802
    """
    Generate a string representation of an object with optional keyword arguments.
    Parameters:
    obj (Any): The object to generate the string representation for.
    **kw: Arbitrary keyword arguments. Supported keywords:
        - 'addr' (bool): If True, include the address of the object
            in the representation.
    Returns:
    str: A string representation of the object, potentially including its
        address and JSON representation.
    """
    from kalib.datastructures import try_json

    kw.setdefault('addr', True)
    msg = Who(obj, **kw)

    if class_of(obj) is not obj:
        msg = f'({msg}):{try_json(obj)}'
    return msg


def Module(obj):  # noqa: N802
    """
    Extract the module name from a given object's fully qualified name.

    Args:
        obj: The object whose module name is to be extracted.

    Returns:
        str: The module name of the given object.
    """
    return Who(obj).rsplit('.', 1)[0]


Who.Name   = partial(Who, full=False)
Who.Is     = Is
Who.Module = Module

#

def sourcefile(obj, **kw):
    kw.setdefault('exclude_self', False)
    kw.setdefault('exclude_stdlib', False)

    for i in iter_inheritance(obj, **kw):
        with suppress(TypeError):
            return getsourcefile(i)


Who.File = sourcefile


def stackoffset(order=None, /, shift=0):
    stack = extract_stack()[:-1]

    def normalize(x):
        if not isinstance(x, bytes | str):
            x = sourcefile(x)
        return str(Path(x))

    @cache
    def is_ignored_frame(x):
        for path in order:
            if x.startswith(path):
                return True
    if order:
        counter = 0
        if isinstance(order, str | bytes):
            order = [order]
        order = set(map(normalize, filter(bool, order)))

        for no, frame in enumerate(reversed(stack)):
            if not is_ignored_frame(frame.filename):
                if counter >= shift:
                    return len(stack) - no -1
                else:
                    counter += 1
    return 0


def stackfilter(line):
    line = line.lstrip()
    if line.startswith('File "<frozen importlib._'):
        return True

    for regex in (
        r'kalib/descriptors\.py", line \d+, in (__get__|call|type_checker)',
    ):
        if search(regex, line):
            return True


def stacktrace(count=None, /, join=True):
    if count is None:
        count = stackoffset(__file__) + 1

    stack = format_stack()

    if count > 0:
        stack = stack[:count]

    elif count <= 0:
        return stack[len(stack) + count -2].rstrip()

    stack = tuple(map(str.strip, filterfalse(stackfilter, stack)))
    return '\n'.join(stack) if join else stack


def about(something):
    try:
        path = sourcefile(something)

    except TypeError:
        try:
            path = getsourcefile(something)
        except TypeError:
            path = None

    mro = get_mro(something, glue=', ')
    std = ('3rd party', 'from stdlib')[is_internal(something)]
    return f'{Who.Is(something)} ({std}) from {path=}, {mro=}'


@cache
def is_imported_module(name):

    with suppress(KeyError):
        return bool(modules[name])

    chunks = name.split('.')
    return sum(
        '.'.join(chunks[:no + 1]) in modules
        for no in range(len(chunks))) >= 2  # noqa: PLR2004

@cache
def get_module_from_path(path):
    def get_path_without_extension(path):
        return splitext(Path(path).absolute().resolve())[0]  # noqa: PTH122

    stem = get_path_without_extension(path)
    for name, module in modules.items():
        with suppress(AttributeError):
            if module.__file__ and stem == get_path_without_extension(module.__file__):
                return name


@cache
def get_subclasses_from_path(cls, path):
    from kalib.importer import required

    if module_path := get_module_from_path(path):
        module = required(module_path)

        result = []
        for name in dir(module):
            if name.startswith('__') or name.endswith('__'):
                continue

            with suppress(Exception):
                if (subclass := getattr(module, name)) and issubclass(subclass, cls):
                    result.append(subclass)

        return tuple(result)


def iter_stack(*args, **kw):
    result = stack()[kw.pop('offset', 0):]
    yield from (map(itemgetter(*args), result) if args else result)


is_awaitable = isawaitable
is_builtin   = isbuiltin
is_class     = isclass
is_coroutine = iscoroutine
is_function  = isfunction
is_method    = ismethod
is_module    = ismodule


Is = (
    namedtuple(
        'Is', 'Class awaitable builtin callable collection coroutine '
        'Builtin Primivite function imported internal '
        'iterable mapping method module primitive')(
    class_of, is_awaitable, is_builtin, is_callable, is_collection,
    is_coroutine, is_from_builtin, is_from_primivite, is_function,
    is_imported_module, is_internal, is_iterable, is_mapping, is_method,
    is_module, is_primitive))
